@@ -91,8 +91,13 @@ enum class ObjectKind {
   Tag,
   Module,
   Instance,
+
+  First = Null,
+  Last = Instance,
 };
 
+static const int kCommandTypeCount = WABT_ENUM_COUNT(ObjectKind);
+
 const char* GetName(Mutability);
 const std::string GetName(ValueType);
 const char* GetName(ExternKind);
@@ -408,6 +413,8 @@ class FreeList {
  public:
   using Index = size_t;
 
+  ~FreeList();
+
   template <typename... Args>
   Index New(Args&&...);
   void Delete(Index);
@@ -421,35 +428,35 @@ class FreeList {
   Index count() const;  // The number of used elements.
 
  private:
-  // TODO: Optimize memory layout? We could probably store all of this
-  // information in one uintptr_t.
-  //
-  // For example, when T is a pointer to an Object (e.g. Store::ObjectList), we
-  // can assume alignment to 4 bytes at least. Given a 32-bit pointer, we can
-  // expect the following layout:
-  //
-  //   nnnnnnnn nnnnnnnn nnnnnnnn nnnnnnn0 f
-  //
-  // where:
-  //   f: "is_free": 0 when the payload is of type T
-  //                 1 when the payload is the index of the next free object
-  //   n: the payload
-  //
-  // When T is a Ref (see Store::RootList below), we'd need to store the
-  // "is_free" bit in most-significant bit instead.
-  //
+  // As for Refs, the free bit is 0x80..0. This bit is never
+  // set for valid Refs, since it would mean more objects
+  // are allocated than the total amount of memory.
+  static const Index refFreeBit = (SIZE_MAX >> 1) + 1;
+
+  // As for Objects, the free bit is 0x1. This bit is never
+  // set for valid Objects, since pointers are aligned to at
+  // least four bytes.
+  static const Index ptrFreeBit = 1;
+  static const int ptrFreeShift = 1;
+
   std::vector<T> list_;
-  std::vector<size_t> free_;
-  std::vector<bool> is_free_;
+  // If free_head_ is zero, there is no free slots in list_,
+  // otherwise free_head_ - 1 represents the first free slot.
+  Index free_head_ = 0;
+  Index free_items_ = 0;
 };
 
 class Store {
  public:
-  using ObjectList = FreeList<std::unique_ptr<Object>>;
+  using ObjectList = FreeList<Object*>;
   using RootList = FreeList<Ref>;
 
   explicit Store(const Features& = Features{});
 
+  Store(const Store&) = delete;
+  Store& operator=(const Store&) = delete;
+  Store& operator=(const Store&&) = delete;
+
   bool IsValid(Ref) const;
   bool HasValueType(Ref, ValueType) const;
   template <typename T>
@@ -481,12 +488,20 @@ class Store {
   template <typename T>
   friend class RefPtr;
 
+  struct GCContext {
+    int call_depth = 0;
+    std::vector<bool> marks;
+    std::vector<size_t> untraced_objects;
+  };
+
+  static const int max_call_depth = 10;
+
   Features features_;
+  GCContext gc_context_;
   // This set contains the currently active Thread objects.
   std::set<Thread*> threads_;
   ObjectList objects_;
   RootList roots_;
-  std::vector<bool> marks_;
 };
 
 template <typename T>
